function [] = asyncLDAsampler(dataset,T,P,seed)
% function [] = asyncLDAsampler(dataset,T,P,seed)
% 
% Runs a simulation of asynchronous distributed sampling for LDA
% (see NIPS 2008 paper).
%
% INPUTS:
%   dataset: 'k' for KOS, 'n' for NIPS
%   T: # of topics
%   P: # of asynchronous processors
%   seed: random seed

% parameters
dataset
T
P
ITER  = 1000;
beta  = 0.01;
alpha = 0.1;
rand('state',sum(100*clock))

% read corpus
if (dataset=='k')
  load docword.kos.train.mat
else
  load docword.nips.train.mat
end

% random initial assignment
z  = floor(T*rand(N,1)) + 1;
wp = zeros(W,T);
dp = zeros(D,T);
for n = 1:N
  wp(w(n),z(n)) = wp(w(n),z(n)) + 1;
  dp(d(n),z(n)) = dp(d(n),z(n)) + 1;
end
ztot = sum(wp,1);

% split over P procs
Dp = D/P;
Nstart = 1;

for p=1:P
  
  Dstart = (p-1)*Dp + 1;
  Dend   =     p*Dp;
  
  Nend   = full(sum(sum(dp(1:Dend,:))));
  
  data(p).w  =  w( Nstart:Nend );
  data(p).d  =  d( Nstart:Nend ) - (Dstart - 1); % local numbering
  data(p).z  =  z( Nstart:Nend );
  data(p).dp = dp( Dstart:Dend, : );
  data(p).nstart = Nstart;
  data(p).nend = Nend;
  data(p).dstart = Dstart;
  data(p).wSum = zeros(1,W);
  % These are the processor's local topic counts (see NIPS paper)
  data2(p).local = zeros(W,T);
  tempSize = length(data(p).w);
  for n=1:tempSize
    data(p).wSum(data(p).w(n)) = data(p).wSum(data(p).w(n)) + 1;
    data2(p).local(data(p).w(n),data(p).z(n)) = data2(p).local(data(p).w(n),data(p).z(n)) + 1;
  end
  
  data(p).wSumIndx = find(data(p).wSum);
  data(p).wSumNnz = data(p).wSum(data(p).wSumIndx);
  data(p).wCount = length(data(p).wSumNnz);
  
  Nstart = Nend + 1;
  
  data2(p).datasets = zeros(1,P);
  data2(p).datasets(p) = 1;
  
end

for p=1:P
  data(p).wp = data2(p).local;
end


% iterate
for iter = 1:ITER
  
  pairing = randperm(P);
  fprintf('iter %d\n', iter);
  
  for p=1:P
    data(p).ztot = sum(data(p).wp);
    % This is the local sampling step on each processor
    [data(p).z, data(p).wp, data(p).dp, tmpLocal] = ...
      gibbsAsyncMex(data(p).z, data(p).wp, data(p).dp, data(p).ztot, data(p).w, data(p).d);
    data2(p).local = tmpLocal;
  end
  
  % Now do the asynchronous pairwise communcation ('gossip-based')
  for pair=1:2:(P-1)
    p1 = pairing(pair); % first processor
    p2 = pairing(pair + 1); % second processor
    
    %fprintf('\tPair: %d, %d\t',p1,p2);
    
    wpGlobal1 = data(p1).wp - data2(p1).local;
    wpGlobal2 = data(p2).wp - data2(p2).local;
    
    % Remove local contribution of other processor   
    if (data2(p1).datasets(p2) == 1)
      tempSize = data(p2).wCount;
      for n = 1:tempSize
        ww = data(p2).wSumIndx(n);
        tts = mhypergeometric(wpGlobal1(ww,:),data(p2).wSumNnz(n));
        wpGlobal1(ww,:) = wpGlobal1(ww,:) - tts;
      end
    else
      data2(p1).datasets(p2) = 1;
    end
    
    if (data2(p2).datasets(p1) == 1)
      tempSize = data(p1).wCount;
      for n = 1:tempSize
        ww = data(p1).wSumIndx(n);
        tts = mhypergeometric(wpGlobal2(ww,:),data(p1).wSumNnz(n));
        wpGlobal2(ww,:) = wpGlobal2(ww,:) - tts;
      end
    else
      data2(p2).datasets(p1) = 1;
    end
    
    % Update the processor's global matrix
    data(p1).wp = wpGlobal1 + data2(p1).local + data2(p2).local;
    data(p2).wp = wpGlobal2 + data2(p1).local + data2(p2).local;

  end
  
  
  % print topics on processor 1
    
  if (mod(iter,1)==0)
    wp = data(1).wp;
    for t=1:T
      [xsort,isort] = sort(-wp(:,t));
      fprintf('[t%d] (%.3f) ', t, ztot(t)/N);
      for i=1:min(8,W)
        fprintf('%s ', word{isort(i)});
      end
      fprintf('\n');
    end
  end

end


